import concurrent.futures
from typing import Callable, Dict, List, Union, Tuple

import numpy as np
import pandas as pd
from pandas.core.frame import DataFrame
from pandas.core.groupby import DataFrameGroupBy
from pandas.io.parsers import TextFileReader

from .array import _DEFAULT_N_WORKERS_, parallelize_iterable
from .code import LabelCoder
import more_itertools


def _apply_wrapper_(dataframe: DataFrame, apply_func, *args, **kwargs):
    return dataframe.apply(apply_func, *args, **kwargs)


def _groupby_apply_wrapper_(list_dataframe: Tuple[DataFrame], apply_func, *args, **kwargs):
    list_dataframe = list(list_dataframe)
    for i, dataframe in enumerate(list_dataframe):
        list_dataframe[i] = apply_func(dataframe, *args, **kwargs)

    concat_dataframe = None
    # concat results of different types
    if isinstance(list_dataframe[0], DataFrame):
        concat_dataframe = pd.concat(list_dataframe, copy=False)
    else:
        parameters = {
            'index': kwargs.get('index', None),
            'columns': kwargs.get('columns', None)
        }
        concat_dataframe = pd.DataFrame.from_records(list_dataframe, **parameters)

    return concat_dataframe


def apply(
        dataframe: DataFrame, func: Callable, *args,
        chunksize: int = None, executor: concurrent.futures.Executor = None,
        n_workers=_DEFAULT_N_WORKERS_, use_process=False,
        **kwargs):

    if chunksize is None:
        chunksize = max(len(dataframe) // n_workers, n_workers)

    def _generator_():
        for i in range(0, len(dataframe), chunksize):
            yield dataframe.iloc[i: i + chunksize]

    list_results = [result for result in parallelize_iterable(
        _generator_(), _apply_wrapper_, func, *args,
        executor=executor, n_workers=n_workers, use_process=use_process, **kwargs
    )]

    dataframe = pd.concat(list_results, copy=False)
    return dataframe


def groupby_apply(
        groups: DataFrameGroupBy, func: Callable, *args,
        chunksize: int = None, executor: concurrent.futures.Executor = None,
        n_workers=_DEFAULT_N_WORKERS_, use_process=False,
        **kwargs):

    # take only the dataframe
    groups = map(lambda g: g[1], groups)

    # chunkify the groups
    if chunksize is None:
        # divide return iterables
        # convert to list as the _groupby_apply_wrapper_ accept list
        # filter empty group which happens if n_groups < n_workers
        groups = more_itertools.divide(n_workers, groups)
        groups = map(list, groups)
        groups = filter(len, groups)
    else:
        groups = more_itertools.chunked(groups, chunksize)

    list_results = [result for result in parallelize_iterable(
        groups, _groupby_apply_wrapper_, func, *args,
        executor=executor, n_workers=n_workers, use_process=use_process, **kwargs
    )]

    dataframe = pd.concat(list_results, copy=False)

    return dataframe


def encode_dataframe(dataframe: DataFrame, columns: Union[List[str], Dict[str, LabelCoder]]):
    if not isinstance(columns, dict):
        columns = {cname: LabelCoder() for cname in columns}

    dataframe = dataframe.assign(**{
        cname: encoder.encode(dataframe[cname].values) for cname, encoder in columns.items()
    })

    return dataframe, columns


def decode_dataframe(dataframe: DataFrame, columns: Dict[str, LabelCoder]):
    dataframe = dataframe.assign(**{
        cname: encoder.decode(dataframe[cname].values) for cname, encoder in columns.items()
    })

    return dataframe


def read_csv_and_encode(filepath, encode_columns: Union[List[str], Dict[str, LabelCoder]], chunksize: int = None, **kargs):
    if chunksize is None:
        dataframe = pd.read_csv(filepath, **kargs)
        dataframe, encode_columns = encode_dataframe(dataframe, encode_columns)
        return dataframe, encode_columns

    dataframe_chunks = []
    with pd.read_csv(filepath, iterator=True, chunksize=chunksize, ** kargs) as reader:
        reader: TextFileReader
        for dataframe in reader:
            dataframe, encode_columns = encode_dataframe(dataframe, encode_columns)
            dataframe_chunks.append(dataframe)

    ignore_index = ('index_col' not in kargs) or (not kargs['index_col'])
    dataframe: DataFrame = pd.concat(dataframe_chunks, ignore_index=ignore_index)
    return dataframe, encode_columns


def decode_and_write_csv(filepath, dataframe: DataFrame, encode_columns: Dict[str, LabelCoder], chunksize: int = None, **kargs):
    if chunksize is None:
        chunksize = len(dataframe)

    dataframe_list = np.array_split(dataframe, range(chunksize, len(dataframe), chunksize))
    dataframe = decode_dataframe(dataframe_list[0], encode_columns)
    dataframe.to_csv(filepath, **kargs)

    for dataframe in dataframe_list[1:]:
        dataframe = decode_dataframe(dataframe, encode_columns)
        dataframe.to_csv(filepath, header=False, mode='a', **kargs)



import unittest
import os

from nure.sync.postgre import PostgreSql
from nure.sync.suffix import ParameterSuffix


class PostgreTestCase(unittest.TestCase):
    def setUp(self) -> None:
        ttl = 10_800
        self.venture_code = 'tw'
        self.db = PostgreSql(
            'tests/credentials/pitbull.json',
            root_path='tests/data/postgre',
            suffix_func=ParameterSuffix({
                'sa_replace': ['venture_code']
            }), ttl=ttl)

    def test_require(self):
        fn = self.db.require('tests/sql/catalogue.sql', sa_replace={'venture_code': self.venture_code})
        self.assertEqual(fn, os.path.join(self.db.root_path, f'catalogue__{self.venture_code}.csv'))



import json
import os
import unittest
from datetime import datetime, timedelta

from nure.sync.bigquery import BigQuery
from nure.sync.suffix import ParameterSuffix


class BigQeuryTestCase(unittest.TestCase):
    def setUp(self) -> None:
        ttl = 10_800
        self.venture_code = 'tw'
        meta_fn = 'tests/credentials/bigquery/metadata.json'
        with open(meta_fn) as fd:
            meta = json.load(fd)

        self.dataset_id = meta[self.venture_code]['dataset_id']
        self.db = BigQuery(
            f'tests/credentials/bigquery/{meta[self.venture_code]["project_id"]}.json',
            root_path='tests/data/bigquery',
            suffix_func=ParameterSuffix({
                're_replace': [R'\$\(date\)', R'\$\(venture_code\)'],
            }),
            ttl=ttl)

    def test_require(self):
        date = (datetime.now() - timedelta(days=1)).strftime('%Y%m%d')
        fn = self.db.require('tests/sql/interaction.sql', re_replace={
            R'\$\(dataset_id\)': self.dataset_id,
            R'\$\(date\)': date,
            R'\$\(venture_code\)': self.venture_code
        })
        self.assertEqual(fn, os.path.join(self.db.root_path, f'interaction__{date}__{self.venture_code}.csv'))



import unittest
from concurrent.futures import ThreadPoolExecutor

import numpy as np
import nure.array
import nure.dataframe
import pandas as pd


def linear(array, slope, inception=0):
    return array * slope + inception


def sum(dataframe, column):
    return pd.Series({
        column: np.sum(dataframe[column])
    })


def numpy_accumulate(generator):
    return np.concatenate(list(generator))


class ParallelTestCase(unittest.TestCase):
    def setUp(self) -> None:
        self.array = np.random.randint(0, 100, size=100000)
        self.df = pd.DataFrame(data={
            'col1': np.random.randint(0, 100, size=100000),
            'col2': np.random.randint(0, 100, size=100000),
        })

    def assert_result(self, a1, a2):
        self.assertEqual(a1.shape, a2.shape)
        self.assertTrue(np.all(a1 == a2))

    def test_array(self):
        slope = 2
        inception = 5
        groundtruth = linear(self.array, slope, inception)
        chunksize = 5000

        def _chunk_generator_():
            for i in range(0, len(self.array), chunksize):
                yield self.array[i: i + chunksize]

        # test with default thread executor
        result = numpy_accumulate(nure.array.parallelize_iterable(
            _chunk_generator_(), linear, slope, inception=inception,
            preserve_order=True, executor=None, use_process=False))

        self.assert_result(result, groundtruth)

        # test with default process executor
        result = numpy_accumulate(nure.array.parallelize_iterable(
            _chunk_generator_(), linear, slope=slope, inception=inception,
            preserve_order=True, executor=None, use_process=True))

        self.assert_result(result, groundtruth)

        # test with defined process executor, specific chunksize
        with ThreadPoolExecutor() as executor:
            result = numpy_accumulate(nure.array.parallelize_iterable(
                _chunk_generator_(), linear, slope, inception=inception,
                preserve_order=True, executor=executor))

        self.assert_result(result, groundtruth)

    def test_dataframe_apply(self):
        slope = 2
        inception = 5
        groundtruth = self.df.apply(linear, slope=slope, inception=inception)
        results = nure.dataframe.apply(
            self.df, linear, slope=slope, inception=inception,
            executor=None, chunksize=None, use_process=False)

        # it does not guarantee the order after groupby
        groundtruth = groundtruth.sort_index()
        results = results.sort_index()

        self.assert_result(results['col1'], groundtruth['col1'])
        self.assert_result(results['col2'], groundtruth['col2'])

    def test_dataframe_groupby_apply(self):
        slope = 2
        inception = 5
        groundtruth = self.df.groupby('col1').apply(linear, slope=slope, inception=inception)
        results = nure.dataframe.groupby_apply(
            self.df.groupby('col1'), linear, slope, inception=inception,
            executor=None, chunksize=None, use_process=False)

        # it does not guarantee the order after groupby
        groundtruth = groundtruth.sort_index()
        results = results.sort_index()

        self.assert_result(results['col1'], groundtruth['col1'])
        self.assert_result(results['col2'], groundtruth['col2'])

    def test_dataframe_groupby_apply_record(self):
        groundtruth = self.df.groupby('col1').apply(sum, 'col2')
        results = nure.dataframe.groupby_apply(
            self.df.groupby('col1'), sum, 'col2',
            executor=None, chunksize=None, use_process=False, retain_order=True)

        self.assert_result(results['col2'].values, groundtruth['col2'].values)



import os
import shutil
import unittest

import numpy as np
import nure.archive


class ArchiveTestCase(unittest.TestCase):
    def setUp(self) -> None:
        self.data = np.random.normal(size=100)
        self.archive_fn = 'tests/data/archive.zip'
        self.root_dir = os.path.splitext(self.archive_fn)[0]

    def tearDown(self) -> None:
        if os.path.isdir(self.root_dir):
            shutil.rmtree(self.root_dir)
        if os.path.isfile(self.archive_fn):
            os.remove(self.archive_fn)

    def test_archive(self):
        archive = nure.archive.Archive(self.archive_fn, self.root_dir)
        data_mod_fn = archive.get('data.npy', mode='m')
        np.save(data_mod_fn, self.data)
        archive.pack()
        self.assertTrue(os.path.isfile(self.archive_fn))

        archive.cleanup()
        archive.ensure_dirs()
        archive.unpack()
        data_org_fn = archive.get('data.npy', mode='o')
        data = np.load(data_org_fn)
        self.assertTrue(np.all(data == self.data))



import unittest
import os

from nure.sync.s3 import S3


class S3TestCase(unittest.TestCase):
    def setUp(self) -> None:
        ttl = 10_800
        self.s3 = S3(
            root_path='tests/data/s3',
            ttl=ttl
        )

    def test_require(self):
        fn = self.s3.require(R's3://zalora.ds-team/LochNess/sku_properties_all.csv')
        self.assertEqual(fn, os.path.join(self.s3.root_path, 'zalora.ds-team/LochNess/sku_properties_all.csv'))

    def test_exist(self):
        self.assertTrue(self.s3.exist(R's3://zalora.ds-team/LochNess/sku_properties_all.csv'))
        self.assertFalse(self.s3.exist(R's3://zalora.ds-team/abc.xyz'))



from typing import Dict, Any, Tuple, List

import numpy as np
import PIL.Image
import PIL.ImageColor
import PIL.ImageDraw
import PIL.ImageFont
from PIL.Image import Image
from PIL.ImageFont import ImageFont

from . import func


def wrap_text(text, font: ImageFont, width):
    w, _ = font.getsize(text)
    max_char = int((width / w) * len(text))
    n_lines = len(text) // max_char + (len(text) % max_char != 0)
    wrapped_text = '\n'.join([text[i * max_char:(i + 1) * max_char] for i in range(n_lines)])
    return wrapped_text


def draw_text(image: Image, text: str, pos=(0, 0), font=None, text_color='white', background_color='black', wrap=False):
    # get font
    if font is None:
        font = PIL.ImageFont.load_default()
    # context
    draw = PIL.ImageDraw.Draw(image)

    # draw background
    if background_color is not None:
        # get text size
        w, h = font.getsize(text)
        # draw rect
        draw.rectangle((pos[0], pos[1], pos[0] + w, pos[1] + h), fill=background_color)
    # texxt wrapping
    if wrap:
        if isinstance / (wrap, int):
            width = wrap
        else:
            width = image.size[0] - pos[0]
        text = wrap_text(text, font, width)

    # draw text
    draw.text(pos, text, fill=text_color, font=font)


def draw_receptive_field(image: Image, size=224, outline=(255, 0, 0)):
    # context
    draw = PIL.ImageDraw.Draw(image)
    # retangle
    image_size = image.size
    left_top = (np.array(image_size) - size) // 2
    right_bottom = left_top + size
    draw.rectangle([*left_top, *right_bottom], outline=outline, width=3)


def draw_rect(image: Image, box=[0, 0, 224, 224], outline=(255, 0, 0), width=2):
    # context
    draw = PIL.ImageDraw.Draw(image)
    # retangle
    draw.rectangle(box, outline=outline, width=width)


__default_colors__ = [PIL.ImageColor.getrgb(c) for c in ['#B22222', '#FF69B4', '#FF6347', '#FFD700', '#ADFF2F', '#40E0D0']]


def draw_bboxes(image: Image, bboxes: np.ndarray, classes: np.ndarray,
                scores: np.ndarray = None, id2name: Dict[Any, str] = None,
                colors=__default_colors__, font: ImageFont = None, copy=True):
    """ draw bounding boxes with their class names on images
    Input:
        image: PIL.Image
        bboxes: [n_boxes, (x1, y1, x2, y2)]
        classes, scores: [n_boxes]
        id2name: dict
    """
    if copy:
        image = image.copy()

    if id2name is not None:
        labels = [id2name.get(cid, 'unknown') for cid in classes]
    else:
        labels = [str(cid) for cid in classes]

    # get font
    if font is None:
        font = PIL.ImageFont.load_default()

    for idx in range(bboxes.shape[0]):
        color = colors[idx % len(colors)]
        x1, y1, x2, y2 = bboxes[idx]

        draw_rect(image, [x1, y1, x2, y2], outline=color)
        text = str(labels[idx])
        if scores is not None:
            text += f' {scores[idx]:.3}'

        _, h = font.getsize(text)
        draw_text(image, text, [x1, y1 - h], font=font, text_color='black', background_color=color)

    return image


def make_grid(images: List[Image], grid: Tuple[int, int] = None, max_size=224, bg_color=(255, 255, 255)):
    """ make a image grid
    Parameters:
        images: array of PIL.Image
            list of image
        grid: 2-tuple of int or int
            (n_col, n_row) or n_col
        max_size: int
            the max dimension of image
        bg_color: 3-tuple, 4-tuple, or str
            background color
    Returns:
        PIL.Image
            grided image
    """
    if grid is None:
        grid = (len(images), 1)
    if isinstance(grid, int):
        grid = (grid, len(images) // grid + (len(images) % grid > 0))

    assert len(grid) == 2
    bg_size = (max_size * grid[0], max_size * grid[1])
    background = PIL.Image.new('RGB', bg_size, bg_color)

    pasted_images = 0
    for row in range(grid[1]):
        for col in range(grid[0]):
            if pasted_images == len(images):
                return background

            left = max_size * col
            top = max_size * row

            nImage = np.array(images[pasted_images])
            nResized = func.resize_keep_ratio(nImage, max_size)
            background.paste(PIL.Image.fromarray(nResized), (left, top))

            pasted_images += 1

    return background


def draw_palette(colors: List[Tuple[int, int, int]], size=64):
    palette = np.empty((size, len(colors) * size, 3), dtype=np.uint8)
    for i in range(len(colors)):
        palette[:, i * size:(i + 1) * size, :] = colors[i]
    return PIL.Image.fromarray(palette)


EDGE_TOP = 0
EDGE_LEFT = 1
EDGE_RIGHT = 2
EDGE_BOTTOM = 3


def draw_text_at_edge(image: Image, text: str, pos=EDGE_BOTTOM,
                      font: ImageFont = None, text_width=100,
                      background_color=0, text_color='white'):

    # get font
    if font is None:
        font = PIL.ImageFont.load_default()

    if pos in [EDGE_TOP, EDGE_BOTTOM]:
        text_width = image.size[0]

    wrapped_text = wrap_text(text, font, text_width)
    _, text_height = font.getsize_multiline(wrapped_text)
    if pos in [EDGE_TOP, EDGE_BOTTOM]:
        image_width = image.size[0]
        image_height = image.size[1] + text_height
    else:
        image_width = image.size[0] + text_width
        image_height = max(image.size[1], text_height)

    image_x = 0
    image_y = 0
    text_x = 0
    text_y = 0
    if pos == EDGE_BOTTOM:
        text_y = image.size[1]
    elif pos == EDGE_LEFT:
        image_x = text_width
    elif pos == EDGE_RIGHT:
        text_x = image.size[0]
    elif pos == EDGE_TOP:
        image_y = text_height

    new_image = PIL.Image.new(image.mode, (image_width, image_height), color=background_color)
    new_image.paste(image, (image_x, image_y))
    # draw text
    draw = PIL.ImageDraw.Draw(new_image)
    draw.text((text_x, text_y), wrapped_text, fill=text_color, font=font)

    return new_image



import math
import os
import struct

import cv2
import numpy as np

from . import annotate


def _apply_mask(matrix, mask, fill_value):
    masked = np.ma.array(matrix, mask=mask, fill_value=fill_value)
    return masked.filled()


def _apply_threshold(matrix, low_value, high_value):
    low_mask = matrix < low_value
    matrix = _apply_mask(matrix, low_mask, low_value)

    high_mask = matrix > high_value
    matrix = _apply_mask(matrix, high_mask, high_value)

    return matrix


def simplest_cb(img, percent):
    assert img.shape[2] == 3
    assert percent > 0 and percent < 100

    half_percent = percent / 200.0

    channels = cv2.split(img)

    out_channels = []
    for channel in channels:
        assert len(channel.shape) == 2
        # find the low and high precentile values (based on the input percentile)
        height, width = channel.shape
        vec_size = width * height
        flat = channel.reshape(vec_size)

        assert len(flat.shape) == 1

        flat = np.sort(flat)

        n_cols = flat.shape[0]

        low_val = flat[math.floor(n_cols * half_percent)]
        high_val = flat[math.ceil(n_cols * (1.0 - half_percent))]

        # saturate below the low percentile and above the high percentile
        thresholded = _apply_threshold(channel, low_val, high_val)
        # scale the channel
        normalized = cv2.normalize(
            thresholded, thresholded.copy(), 0, 255, cv2.NORM_MINMAX)
        out_channels.append(normalized)

    return cv2.merge(out_channels)


def unpad(image):
    pad_color = image[0, 0]
    # top
    top = 0
    while True:
        if not np.all(image[top] == pad_color):
            break
        top += 1

    # left
    left = 0
    while True:
        if not np.all(image[:, left] == pad_color):
            break
        left += 1

    # top
    bottom = -1
    while True:
        if not np.all(image[bottom] == pad_color):
            break
        top -= 1

    # left
    right = -1
    while True:
        if not np.all(image[:, right] == pad_color):
            break
        right -= 1

    return image[top:bottom, left:right, :]


def resize_keep_ratio(image, max_size, interpolation=cv2.INTER_AREA):
    h, w = np.shape(image)[:2]

    if h < w:
        h = int(max_size * h / w)
        w = max_size
    else:
        w = int(max_size * w / h)
        h = max_size

    image = cv2.resize(image, (w, h), interpolation=interpolation)
    return image


def create_mosaic(nImages, list_gt_boxes, list_gt_classes):
    """ create a mosaic image with 4 images and transform their groundtruth boxes
    """
    # the dtype
    image_dtype = nImages[0].dtype
    # compute center and width, hiehgt of combined image
    cy = max(nImages[0].shape[0], nImages[1].shape[0])
    height = cy + max(nImages[2].shape[0], nImages[3].shape[0])
    cx = max(nImages[0].shape[1], nImages[2].shape[1])
    width = cx + max(nImages[1].shape[1], nImages[3].shape[1])

    # paste 4 images into
    nBig = np.zeros((height, width, 3), dtype=image_dtype)
    nBig[cy - nImages[0].shape[0]:cy, cx - nImages[0].shape[1]:cx] = nImages[0]
    nBig[cy - nImages[1].shape[0]:cy, cx:cx + nImages[1].shape[1]] = nImages[1]
    nBig[cy:cy + nImages[2].shape[0], cx - nImages[2].shape[1]:cx] = nImages[2]
    nBig[cy:cy + nImages[3].shape[0], cx:cx + nImages[3].shape[1]] = nImages[3]

    # translate the bboxes according to their possition
    translated_gt_boxes = [
        annotate.translate_boxes(list_gt_boxes[0], (cy - nImages[0].shape[0], cx - nImages[0].shape[1])),
        annotate.translate_boxes(list_gt_boxes[1], (cy - nImages[1].shape[0], cx)),
        annotate.translate_boxes(list_gt_boxes[2], (cy, cx - nImages[2].shape[1])),
        annotate.translate_boxes(list_gt_boxes[3], (cy, cx))
    ]

    # merge bboxes and classes
    merged_gt_boxes = np.concatenate(translated_gt_boxes, axis=0).reshape(-1, 4)
    merged_gt_classes = np.concatenate(list_gt_classes, axis=0)

    return nBig, merged_gt_boxes, merged_gt_classes


def is_joeg(bytes):
    return bytes[:2] == b'\xff\xd8' and bytes[-2:] == b'\xff\xd9'


def get_image_size(file_path):
    """
    Return (width, height) for a given img file content - no external
    dependencies except the os and struct modules from core
    https://raw.githubusercontent.com/scardine/image_size/master/get_image_size.py
    """
    size = os.path.getsize(file_path)

    with open(file_path) as input:
        height = -1
        width = -1
        data = input.read(25)

        if (size >= 10) and data[:6] in ('GIF87a', 'GIF89a'):
            # GIFs
            w, h = struct.unpack("<HH", data[6:10])
            width = int(w)
            height = int(h)
        elif ((size >= 24) and data.startswith('\211PNG\r\n\032\n') and (data[12:16] == 'IHDR')):
            # PNGs
            w, h = struct.unpack(">LL", data[16:24])
            width = int(w)
            height = int(h)
        elif (size >= 16) and data.startswith('\211PNG\r\n\032\n'):
            # older PNGs?
            w, h = struct.unpack(">LL", data[8:16])
            width = int(w)
            height = int(h)
        elif (size >= 2) and data.startswith('\377\330'):
            # JPEG
            msg = " raised while trying to decode as JPEG."
            input.seek(0)
            input.read(2)
            b = input.read(1)
            try:
                w = 0
                h = 0
                while (b and ord(b) != 0xDA):
                    while (ord(b) != 0xFF):
                        b = input.read(1)
                    while (ord(b) == 0xFF):
                        b = input.read(1)
                    if (ord(b) >= 0xC0 and ord(b) <= 0xC3):
                        input.read(3)
                        h, w = struct.unpack(">HH", input.read(4))
                        break
                    else:
                        input.read(int(struct.unpack(">H", input.read(2))[0]) - 2)
                    b = input.read(1)
                width = int(w)
                height = int(h)
            except struct.error:
                raise Exception("StructError" + msg)
            except ValueError:
                raise ValueError("ValueError" + msg)
            except Exception as e:
                raise Exception(e.__class__.__name__ + msg)
        else:
            raise Exception(
                "Sorry, don't know how to get information from this file."
            )

    return width, height



import os
import unittest

from nure.sync.googlesheet import GoogleSheet


class GoogleSheetTestCase(unittest.TestCase):
    def setUp(self) -> None:
        ttl = 10_800
        self.googlesheet = GoogleSheet(
            'tests/credentials/googlesheet_key.pickle',
            'tests/credentials/googlesheet_secret.json',
            root_path='tests/data/googlesheet',
            ttl=ttl)

    def test_require(self):
        key = {
            'spreadsheet_id': '1CjuMlcz2LT6FkxV21zXyITRfBFblSGVqEIB0_3mZX3E',
            'range': 'subcats_exclude!A2:C'
        }

        fn = self.googlesheet.require(key, header=['country', 'segment', 'sub_cat_type'])
        self.assertEqual(fn, os.path.join(self.googlesheet.root_path, key['spreadsheet_id'], f'{key["range"].split("!")[0]}.csv'))



import numpy as np


class LabelCoder:
    def __init__(self) -> None:
        self._code_dict = {
            None: 0
        }

    def _to_1d_(array):
        # convert to array and reshape to 1-d array
        np_arr = np.asarray(array)
        original_shape = np.shape(np_arr)

        np_arr = np.ravel(np_arr)
        return np_arr, original_shape

    def encode(self, array, return_value=True, expand=True):
        np_arr, original_shape = LabelCoder._to_1d_(array)

        if expand:
            start_code = len(self._code_dict)
            updating_code = set(np_arr) - set(self._code_dict)
            updating_code = {value: start_code + index for index, value in enumerate(updating_code)}
            self._code_dict.update(updating_code)

        if not return_value:
            return

        np_encoded = np.asarray([self._code_dict.get(value, 0) for value in np_arr])
        np_encoded = np.reshape(np_encoded, original_shape)
        return np_encoded

    def decode(self, array):
        np_arr, original_shape = LabelCoder._to_1d_(array)
        # since Python 3.7, the order of inserting to dict is preserved.
        np_codes = np.asarray(list(self._code_dict.keys()))
        # decode and reshape
        np_decoded = np_codes[np_arr]
        np_decoded = np.reshape(np_decoded, original_shape)
        return np_decoded



import math

import numpy as np

ESP = np.finfo(np.float32).eps


def to_ndarray(a):
    if not isinstance(a, np.ndarray):
        a = np.array(a)

    return a


def polygon2bbox(polygon):
    xys = np.reshape(polygon, (-1, 2))
    x1, y1 = np.min(xys, axis=0)
    x2, y2 = np.max(xys, axis=0)

    return x1, y1, x2 - x1, y2 - y1


def normalize_bboxes(bboxes, image_size):
    """ normalize the bboxes to [0, 1]
    Args:
        bboxes: array of shape [N, 4]
            each bounding box contains 4 values likes (ymin, xmin, ymax, xmax)
        image_size: tuple of 2 elements
            image size as (width, height)
    """
    bboxes = to_ndarray(bboxes)
    image_size = to_ndarray(image_size)

    width, height = image_size[..., 0], image_size[..., 1]

    normalized = np.stack((
        bboxes[..., 0] / height, bboxes[..., 1] / width,
        bboxes[..., 2] / height, bboxes[..., 3] / width), axis=-1)

    return normalized


def denormalize_bboxes(bboxes, image_size):
    """ inverse the action of normalize_bboxes
    Args:
        bboxes: array of shape [N, 4]
            each bounding box contains 4 values likes (ymin, xmin, ymax, xmax)
        image_size: tuple of 2 elements
            image size as (width, height)
    """
    image_size = to_ndarray(image_size)
    return normalize_bboxes(bboxes, 1 / image_size)


def calibrate_bboxes(boxes, image_size):
    """ calibrate the bounding boxes when the image is padded to sqaure
    """
    boxes = to_ndarray(boxes)
    image_size = to_ndarray(image_size)

    ymin, xmin, ymax, xmax = boxes[..., 0], boxes[..., 1], boxes[..., 2], boxes[..., 3]
    width, height = image_size[..., 0], image_size[..., 1]

    if width < height:
        ratio = width / height
        xmin = xmin * ratio
        xmax = xmax * ratio
    else:
        ratio = height / width
        ymin = ymin * ratio
        ymax = ymax * ratio

    return np.stack((ymin, xmin, ymax, xmax), axis=-1)


def xywh2yxyx(boxes):
    xmin, ymin, width, height = boxes[..., 0], boxes[..., 1], boxes[..., 2], boxes[..., 3]

    ymax = ymin + height
    xmax = xmin + width

    return np.stack((ymin, xmin, ymax, xmax), axis=-1)


def xywh2xyxy(boxes):
    xmin, ymin, width, height = boxes[..., 0], boxes[..., 1], boxes[..., 2], boxes[..., 3]

    ymax = ymin + height
    xmax = xmin + width

    return np.stack((xmin, ymin, xmax, ymax), axis=-1)


def yxyx2xywh(boxes):
    ymin, xmin, ymax, xmax = boxes[..., 0], boxes[..., 1], boxes[..., 2], boxes[..., 3]

    height = ymax - ymin
    width = xmax - xmin

    return np.stack((xmin, ymin, width, height), axis=-1)


def xyxy2yxyx(boxes):
    xmin, ymin, xmax, ymax = boxes[..., 0], boxes[..., 1], boxes[..., 2], boxes[..., 3]
    return np.stack((ymin, xmin, ymax, xmax), axis=-1)


def yxyx2xyxy(boxes):
    ymin, xmin, ymax, xmax = boxes[..., 0], boxes[..., 1], boxes[..., 2], boxes[..., 3]
    return np.stack((xmin, ymin, xmax, ymax), axis=-1)


def boxes_area(boxes):
    ymin, xmin, ymax, xmax = boxes[..., 0], boxes[..., 1], boxes[..., 2], boxes[..., 3]
    areas = (xmax - xmin) * (ymax - ymin)
    return areas


def intersect_over_union(boxA, boxB):
    boxA = to_ndarray(boxA)
    boxB = to_ndarray(boxB)
    # determine the (x, y)-coordinates of the intersection rectangle
    y1 = np.maximum(boxA[..., 0], boxB[..., 0])
    x1 = np.maximum(boxA[..., 1], boxB[..., 1])
    y2 = np.minimum(boxA[..., 2], boxB[..., 2])
    x2 = np.minimum(boxA[..., 3], boxB[..., 3])

    # compute the area of intersection rectangle
    inter_area = np.maximum(0, x2 - x1) * np.maximum(0, y2 - y1)
    boxA_area = boxes_area(boxA)
    boxB_area = boxes_area(boxB)
    iou = inter_area / (boxA_area + boxB_area - inter_area)

    return iou


def intersect_over_b(boxA, boxB):
    boxA = to_ndarray(boxA)
    boxB = to_ndarray(boxB)
    # determine the (x, y)-coordinates of the intersection rectangle
    y1 = np.maximum(boxA[..., 0], boxB[..., 0])
    x1 = np.maximum(boxA[..., 1], boxB[..., 1])
    y2 = np.minimum(boxA[..., 2], boxB[..., 2])
    x2 = np.minimum(boxA[..., 3], boxB[..., 3])

    # compute the area of intersection rectangle
    interArea = np.maximum(0, x2 - x1) * np.maximum(0, y2 - y1)
    boxBArea = (boxB[..., 2] - boxB[..., 0]) * (boxB[..., 3] - boxB[..., 1])
    iob = interArea / (boxBArea + ESP)

    return iob


def clip_boxes_out_of_image(boxes, image_size, threshold=0.4):
    """Clip off parts of the BB box that are outside of the image in-place.
    Parameters
    ----------
    boxes: tuple[N,4]
        bounding box [ymin, xmin, ymax, xmax]
    image_size:
        size of image [height, width]
    Returns
    -------
    tuple[N',4]
        Bounding box, clipped to fall within the image dimensions.
        The object may have been modified in-place.
    """
    boxes = to_ndarray(boxes)
    image_size = to_ndarray(image_size)
    ymin, xmin, ymax, xmax = boxes[..., 0], boxes[..., 1], boxes[..., 2], boxes[..., 3]
    height, width = image_size[..., 0], image_size[..., 1]

    xmin = np.clip(xmin, 0, width - ESP)
    xmax = np.clip(xmax, 0, width - ESP)
    ymin = np.clip(ymin, 0, height - ESP)
    ymax = np.clip(ymax, 0, height - ESP)

    new_boxes = np.stack((ymin, xmin, ymax, xmax), axis=-1)
    if threshold is None:
        return new_boxes, np.ones_like(new_boxes)

    old_area = boxes_area(boxes)
    new_area = boxes_area(new_boxes)
    mask = new_area / old_area > threshold

    return new_boxes, mask


def translate_boxes(boxes, yx):
    """ translate boxes [yxyx] with offset yx
    """
    yx = to_ndarray(yx)
    offset_y, offset_x = yx[..., 0], yx[..., 1]
    ymin, xmin, ymax, xmax = boxes[..., 0], boxes[..., 1], boxes[..., 2], boxes[..., 3]
    xmin = xmin + offset_x
    xmax = xmax + offset_x
    ymin = ymin + offset_y
    ymax = ymax + offset_y

    return np.stack((ymin, xmin, ymax, xmax), axis=-1)


def expand_bounding_boox(image_sizes, boxes, margin) -> np.ndarray:
    """ expand the bounding box evenly all direction by margin percent
    Paramters:
        image_size: Tuple[int, int]
            height, width of image
        bbox: Tuple[int, int, int, int]
            original bounding box. Format in ymin, xmin, ymax, xmax
        margin: float
            the proportion of expansion of each edge. Ex: 0.1 is 10%
    Returns:
        Tuple[int, int, int, int]
            expanded bounding box accounting for image size
    """
    image_sizes = to_ndarray(image_sizes)
    boxes = to_ndarray(boxes)

    ymin, xmin, ymax, xmax = boxes[..., 0], boxes[..., 1], boxes[..., 2], boxes[..., 3]
    w = xmax - xmin
    h = ymax - ymin
    xmin = np.maximum(0, xmin - margin * w)
    ymin = np.maximum(0, ymin - margin * h)
    xmax = np.minimum(image_sizes[..., 1], xmax + margin * w)
    ymax = np.minimum(image_sizes[..., 0], ymax + margin * h)

    return np.stack((ymin, xmin, ymax, xmax), axis=-1)


def cut_instance(image_size, size, bboxes=None, classes=None, overlap=0.1, acceptable_area=0.4):
    """cut image into multiple patch
    Args:
        image_size (tuple(int, int)): original image size (height, width)
        size (int): image patch size
        bboxes (list(4-int,), optional): list of bounding boxes. Defaults to None.
        classes (list(4-int,)), optional): class of boxes. Defaults to None.
        overlap (float, optional): overlap percentage. Defaults to 0.1.
        acceptable_area (float, optional): any cutted boxes whose area is smaller than original one is remove. Defaults to 0.4.
    Returns:
        resolution, new_rects, boxes_list, classes_list
    """
    height, width = image_size
    rows = math.ceil((height - overlap * size) / (size - overlap * size))
    cols = math.ceil((width - overlap * size) / (size - overlap * size))
    resolution = [rows, cols]

    ymin = np.arange(rows) * size * (1 - overlap)
    xmin = np.arange(cols) * size * (1 - overlap)
    ymax = ymin + size
    xmax = xmin + size
    ymax[-1] = height
    xmax[-1] = width

    xx, yy = np.meshgrid(np.arange(cols), np.arange(rows))
    new_rects = np.stack([
        np.take(ymin, yy),
        np.take(xmin, xx),
        np.take(ymax, yy),
        np.take(xmax, xx),
    ], axis=-1).reshape((-1, 4))

    if bboxes is None:
        return resolution, new_rects
    assert classes is not None

    # shape [1, N, 4]
    rects = new_rects.reshape((1, -1, 4))

    # shape [B, 1, 4]
    bboxes = bboxes.reshape((-1, 1, 4))

    # propagate together and get [B, N, 4]
    boxes_to_rects = np.stack([
        bboxes[..., 0] - rects[..., 0],
        bboxes[..., 1] - rects[..., 1],
        bboxes[..., 2] - rects[..., 0],
        bboxes[..., 3] - rects[..., 1]
    ], axis=-1)

    # shape [1, N, 2]
    rects_size = np.stack([
        rects[..., 2] - rects[..., 0],
        rects[..., 3] - rects[..., 1],
    ], axis=-1)

    # clip boxes respect to rects
    # shape [B, N, 4] and [B, N]
    new_boxes, mask = clip_boxes_out_of_image(boxes_to_rects, rects_size, threshold=acceptable_area)

    boxes_list = [new_boxes[mask[:, i], i] for i in range(new_rects.shape[0])]
    classes_list = [classes[mask[:, i]] for i in range(new_rects.shape[0])]

    return resolution, new_rects, boxes_list, classes_list


def _non_maxima_suppression_(boxes, scores=None, threshold=0.7):
    inverted_indices = None
    if scores is not None:
        indices = np.argsort(scores)[::-1]
        boxes = boxes[indices]
        inverted_indices = np.argsort(indices)

    mask = np.ones(boxes.shape[0], dtype=np.bool)
    for i in range(boxes.shape[0]):
        ious = intersect_over_union(boxes[i], boxes[i + 1:])
        suppressed_indices = np.where(ious > threshold)[0] + i + 1
        mask[suppressed_indices] = False

    if scores is None:
        return mask
    return mask[inverted_indices]


def non_maxima_suppression(boxes, scores=None, classes=None, threshold=0.7):
    if classes is None:
        return _non_maxima_suppression_(boxes, scores=scores, threshold=threshold)

    class_uniques = np.unique(classes)
    sup_mask = np.ones(boxes.shape[0], dtype=np.bool)
    for uclass in class_uniques:
        cmask = classes == uclass
        sup_cmask = _non_maxima_suppression_(
            boxes[cmask],
            scores=None if scores is None else scores[cmask], threshold=threshold)
        sup_mask[cmask] = sup_cmask

    return sup_mask



import os
import tempfile
import unittest

import numpy as np
import nure.code
import nure.dataframe
import pandas as pd


class CoderTestCase(unittest.TestCase):
    def setUp(self) -> None:
        self.np_chars = np.array([chr(c) for c in range(ord('A'), ord('Z'))])
        self.df = pd.DataFrame(data={
            'col1': np.random.choice(self.np_chars, size=10000, replace=True),
            'col2': np.random.choice(self.np_chars, size=10000, replace=True)
        })

        self.df.to_csv('tests/data/coder.csv', index=False)

    def test_coder(self):
        coder = nure.code.LabelCoder()
        # check return_value False
        encoded = coder.encode(['A', 'B', 'C'], return_value=False, expand=True)
        self.assertIsNone(encoded)
        self.assertEqual(
            set(coder._code_dict),
            set([None, 'A', 'B', 'C'])
        )

        # check expand False
        encoded = coder.encode(['A', 'B', 'C', 'D', 1, None], return_value=True, expand=False)
        self.assertEqual(
            set(coder._code_dict),
            set([None, 'A', 'B', 'C'])
        )

        # check encoded function
        decoded = coder.decode(encoded)
        self.assertTrue(np.all(decoded == ['A', 'B', 'C', None, None, None]))

        # check expand True
        encoded = coder.encode(['E'], return_value=True, expand=True)
        self.assertEqual(
            set(coder._code_dict),
            set([None, 'A', 'B', 'C', 'E'])
        )

    def test_dataframe(self):
        df1, coders1 = nure.dataframe.read_csv_and_encode('tests/data/coder.csv', encode_columns=['col1', 'col2'], chunksize=None)
        df2, coders2 = nure.dataframe.read_csv_and_encode('tests/data/coder.csv', encode_columns=['col1', 'col2'], chunksize=100)
        # check has encoders
        self.assertTrue('col1' in coders1 and 'col2' in coders1)
        # check encoding correctly
        self.assertEqual(
            set(coders1['col1']._code_dict),
            set(coders2['col1']._code_dict)
        )

        # check shapes
        self.assertEqual(self.df.shape, df1.shape)
        self.assertEqual(self.df.shape, df2.shape)

        # check save function
        with tempfile.TemporaryDirectory() as tmpdir:
            out1_fn = os.path.join(tmpdir, 'out1.csv')
            nure.dataframe.decode_and_write_csv(out1_fn, df1, coders1, chunksize=None, index=False)
            self.assertTrue(os.path.exists(out1_fn))

            out2_fn = os.path.join(tmpdir, 'out2.csv')
            nure.dataframe.decode_and_write_csv(out2_fn, df2, coders2, chunksize=100, index=False)
            self.assertTrue(os.path.exists(out2_fn))

            df1 = pd.read_csv(out1_fn)
            df2 = pd.read_csv(out2_fn)
            for col in df1.columns:
                self.assertTrue(np.all(df1[col] == df2[col]))

    def tearDown(self) -> None:
        os.remove('tests/data/coder.csv')
